#!/usr/bin/env python
# -*- coding: utf-8 -*-*#
"""
Dialog for selecting channels from a station
Lloyd Carothers
IRIS/PASSCAL
"""

import os

from PySide6.QtCore import Qt
from PySide6.QtGui import QFontMetrics
from PySide6.QtUiTools import loadUiType
from PySide6.QtWidgets import (
    QTableWidgetItem,
    QAbstractItemView,
    QHeaderView)

from nexus.obspy_improved import utc_to_str


def load_ui(filename: str) -> object:
    """
    Helper function.
    Loads a ui file relative to this source file.

    Args:
        filename (str): Name of ui file to load.

    Raises:
        e: Exception encountered creating path
           from given filename.

    Returns:
        object: Returns Qt object from ui file.
    """
    path = os.path.join(os.path.dirname(__file__), filename)
    try:
        ret = loadUiType(path)
    except Exception as e:
        print(e)
        raise e
    return ret


class ChannelSelectDialog(*load_ui('ChannelSelectDialog.ui')):
    """
    Dialog for selecting channels from a station.

    Inherits from the generated UI class and QDialog. Provides a user interface
    to display channel information and allow users to select channels.
    """
    def __init__(self, station, parent=None):
        """
        Initialize the ChannelSelectDialog.

        Args:
            parent (Optional[QDialog]): The parent widget. Defaults to None.
        """
        self.station = station
        self.chan_map = {}
        self.selected_channels = []
        super().__init__(parent)
        self.setupUi()

    def accept(self) -> None:
        rows = set(item.row() for item in self.tableWidget.selectedItems())
        for row in rows:
            self.selected_channels.append(self.chan_map[row])
        super().accept()

    def sort_table(self, column: int) -> None:
        """
        Sorts the channel selection table based on column
        header selected by user.

        Args:
            column (int): The index of the user selected
                          column.
        """
        row_map = {
            0: 'location_code',
            1: 'code',
            2: 'sample_rate',
            3: 'start_date',
            4: 'end_date'
        }
        order = self.tableWidget.horizontalHeader().sortIndicatorOrder()
        descending = order != Qt.AscendingOrder
        self.chan_map = {
            ind: v for ind, (_, v) in enumerate(
                sorted(
                    self.chan_map.items(),
                    key=lambda item: getattr(item[1], row_map[column]),
                    reverse=descending
                )
            )
        }

    def setupUi(self) -> None:
        super().setupUi(self)
        self.setWindowTitle('Select channels')
        self.label.setText(f'Station: {self.station.code}')
        self.tableWidget.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.tableWidget.setRowCount(len(self.station))
        self.tableWidget.horizontalHeader().sectionClicked.connect(
            self.sort_table
        )
        # Add channels
        for row, chan in enumerate(self.station):
            for col, value in enumerate(
                    (f"{chan.location_code:>2}",
                     f"{chan.code:<3}",
                     f"{chan.sample_rate:>6.1f}",
                     f"{utc_to_str(chan.start_date)}",
                     f"{utc_to_str(chan.end_date)}",
                     )):
                item = QTableWidgetItem()
                item.setText(value)
                # shorten start/end time
                if col in (3, 4):
                    # full time as tooltip
                    item.setToolTip(value)
                    # rounded time in table
                    value = QFontMetrics(self.tableWidget.font()).elidedText(
                        value,
                        Qt.ElideRight,
                        100
                    )
                    item.setText(value)
                self.tableWidget.setItem(row, col, item)
            self.chan_map[row] = chan
        assert True
        self.tableWidget.resizeColumnsToContents()
        self.tableWidget.horizontalHeader().setSectionResizeMode(
            QHeaderView.Fixed
        )
